from .types import Square, Uci

def uci_as_square(uci : str) -> int: ...
def square_file(sq : Square) -> int: ...
def square_rank(sq : Square) -> int: ...
def square_distance(sq1 : Square, sq2 : Square) -> int: ...
def square_mirror(sq : Square, is_vertical : bool = True) -> int: ...